const abi = require('web3-eth-abi');

const Definitions = {
  Registry: require('../build/definitions/Registry.json'),
  Governance: require('../build/definitions/Governance.json'),
  DomainToken: require('../build/definitions/DomainToken.json'),
  RegistryToken: require('../build/definitions/RegistryToken.json'),
  GovernanceToken: require('../build/definitions/GovernanceToken.json')
};

class Contract {
  constructor(name, network={}) {
    this.definition = Definitions[name];
    this.address = (network[name] || '').toLowerCase();
    this.name = name;
  }

  sigFor(name, type) {
    type = type || name.type;
    name = name.signature || (name.topics || [])[0] || name;

    if (typeof name === 'string' && name.slice(0,2) == '0x') {
      return name;
    } else if (type === 'function') {
      return abi.encodeFunctionSignature(name);
    } else if (type === 'event') {
      return abi.encodeEventSignature(name);
    } else {
      return (
        this.abiFor(this.sigFor(name, 'function')) ||
        this.abiFor(this.sigFor(name, 'event')) ||
        {}
      ).signature;
    }
  }

  abiFor(sig) {
    return this.definition.abi[this.definition.signatures[sig]];
  }

  abi(name, type) {
    if (!name) { return this.definition.abi; }
    else { return this.abiFor(this.sigFor(name, type)); }
  }

  encode(name, args=[], opts={}) {
    let data = abi.encodeFunctionCall(
      this.abi(name, 'function'),
      args
    );

    return {to: this.address, data, ...opts};
  }

  decode(name, data, topics) {
    data = data || name.data;
    topics = topics || name.topics;

    let definition = this.abi(name) || {};
    if (definition.type === 'function') {
      return this.decodeFunction(name, data);
    } else if (definition.type === 'event') {
      return this.decodeEvent(data, topics, definition);
    }
    return false;
  }

  decodeFunction(name, data) {
    let definition = this.abi(name);
    return abi.decodeParameters(definition.outputs, data);
  }

  decodeTransaction(input) {
    let sig = input.slice(0, 10);
    let data = '0x' + input.slice(10);
    let definition = this.abiFor(sig) || {};

    let decoded = {};
    let values = abi.decodeParameters(definition.inputs, data);
    definition.inputs.forEach((input) => {
      decoded[input.name] = values[input.name];
    });

    let types = definition.inputs.map((input) => { return input.type; });
    let name = definition.name + '(' + types.join(',') + ')';

    return {name, decoded, sig};
  }

  decodeEvent(data, topics, definition) {
    data = data.data || data;
    topics = topics || data.topics || [];
    definition = definition || this.abiFor(topics[0]);

    if (!definition.anonymous) { topics.shift(); }

    let decoded = {};
    let values = abi.decodeLog(definition.inputs, data, topics);
    definition.inputs.forEach((input) => {
      decoded[input.name] = values[input.name];
    });

    return decoded;
  }
}

Contract.Definitions = Definitions;

module.exports = Contract;
